/*
 *  aplay_engine.c	(C) 1999 by Andy Lo A foe
 *  Based on aplay Copyright (c) by Jaroslav Kysela <perex@jcu.cz>
 *  Based on vplay program by Michael Beck
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <malloc.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <netinet/in.h>
#include <glib.h>
#include "formats.h"
#include "../input_plugin.h"

#define APLAY_FRAME_SIZE	4608

#define FORMAT_DEFAULT		-1
#define FORMAT_RAW		0
#define FORMAT_VOC		1
#define FORMAT_WAVE		2
#define FORMAT_AU		3

typedef struct _the_format
{
	int rate;
	int channels;
	int bits;
} the_format;

					
struct aplay_local_data
{
	char wav_name[FILENAME_MAX+1];
	int count;
	FILE *wav_fd;
	int header_size;
	the_format format;
};


static void init_aplay()
{
	return;
}


/*
 * test, if it's a .WAV file, 0 if ok (and set the speed, stereo etc.)
 *                            < 0 if not
 */
static int test_wavefile(struct aplay_local_data *data, void *buffer)
{
	WaveHeader *wp = (WaveHeader *)buffer;

	if (wp->main_chunk == WAV_RIFF && wp->chunk_type == WAV_WAVE &&
	    wp->sub_chunk == WAV_FMT && wp->data_chunk == WAV_DATA) {
		if (wp->format != WAV_PCM_CODE) {
			fprintf(stderr, "APLAY: can't play not PCM-coded WAVE-files\n");
			return -1;
		}
		if (wp->modus < 1 || wp->modus > 32) {
			fprintf(stderr, "APLAY: can't play WAVE-files with %d tracks\n",
				wp->modus);
			return -1;
		}
		switch (wp->bit_p_spl) {
		case 8:
			//fprintf(stderr, "APLAY: no support for 8-bit  yet\n");
			//return -1;
		case 16:
			// Humm, no nothing :)
			break;
		default:
			fprintf(stderr, "APLAY: can't play WAVE-files with sample %d bits wide\n",
				wp->bit_p_spl);
		}
		data->format.rate = wp->sample_fq;
		data->format.channels = wp->modus;
		data->format.bits = wp->bit_p_spl;
		data->count = wp->data_length;
		return 0;
	}
	return -1;
}


static int aplay_open(input_object *obj, char *name)
{
	static int aplay_init_done = 0;	
	char audiobuf[4096];
	if (!obj)
			return 0;
	pthread_mutex_lock(&obj->object_mutex);

	obj->local_data = malloc(sizeof(struct aplay_local_data));
	if (!obj->local_data) {
			pthread_mutex_unlock(&obj->object_mutex);
			return 0;
	}
	struct aplay_local_data *data = (struct aplay_local_data *)obj->local_data;
	if (!aplay_init_done) {
			init_aplay();
			aplay_init_done = 1;
	}	
	if (!name || !strcmp(name, "-")) {
		printf("APLAY: Uhm, we don't support stdin\n");
		pthread_mutex_unlock(&obj->object_mutex);
		return 0;
	} else {
		if ((data->wav_fd = fopen(name, "r")) == NULL) {
			perror(name);
			pthread_mutex_unlock(&obj->object_mutex);
			return 0;
		}
	}

	/* read the file header */
	if (fread(audiobuf, sizeof(WaveHeader), 1, data->wav_fd) != 1) {
		fprintf(stderr, "APLAY: read error");
		fclose(data->wav_fd);
		pthread_mutex_unlock(&obj->object_mutex);
		return 0;
	}
	if (test_wavefile((struct aplay_local_data *)obj->local_data, audiobuf) >= 0) {
		// Extract the filename
		char *ptr = strrchr(name, '/');
		if (ptr)
			ptr++;
		else
			ptr = name;
		if (strlen(ptr) > FILENAME_MAX) {
			strncpy(data->wav_name, ptr, FILENAME_MAX-1);
			data->wav_name[FILENAME_MAX-1] = 0;
		} else 
			strcpy(data->wav_name, ptr);
		data->header_size = sizeof(WaveHeader);
		pthread_mutex_unlock(&obj->object_mutex);
		return 1;
	}
	if (data->wav_fd != NULL)
		fclose(data->wav_fd);
	pthread_mutex_unlock(&obj->object_mutex);
	return 0;
}


void aplay_close(input_object *obj)
{
	if (!obj)
			return;
	pthread_mutex_lock(&obj->object_mutex);
	struct aplay_local_data *data = (struct aplay_local_data *)obj->local_data;
	if (!data) {
			pthread_mutex_unlock(&obj->object_mutex);
			return;
	}		
	if (data->wav_fd != NULL)
			fclose(data->wav_fd);
	data->wav_fd =NULL;
	if (obj->local_data)
			free(obj->local_data);
	obj->local_data = NULL;
	pthread_mutex_unlock(&obj->object_mutex);
}

static int aplay_play_frame(input_object *obj)
{
	char audiobuf[8192];
	char tmpbuf[8192];
	if (!obj)
			return 0;
	pthread_mutex_lock(&obj->object_mutex);
	struct aplay_local_data *data = (struct aplay_local_data *)obj->local_data;
	if (!data) {
			pthread_mutex_unlock(&obj->object_mutex);
			return 0;
	}

	// Take care of mono streams
	if (data->format.channels == 1) { // mono, so double 
			if (data->format.bits == 8) { // 8 bit
				if (fread(tmpbuf, APLAY_FRAME_SIZE >> 2, 1, data->wav_fd) != 1) {
					pthread_mutex_unlock(&obj->object_mutex);	
					return 0;
				}
				guint8 *s = (guint8 *)tmpbuf;
				gint16 *d = (gint16 *)audiobuf;
				gint16 c;
				for (int i=0; i < APLAY_FRAME_SIZE; i+=4) {
					*d++ = c = ((*s ^ 0x80) << 8) | *s++;
					*d++ = c;
				}
			} else { // 16 bit					
				if (fread(tmpbuf, APLAY_FRAME_SIZE >> 1, 1, data->wav_fd) != 1) {
						pthread_mutex_unlock(&obj->object_mutex);
						return 0;
				}
				gint16 *s = (gint16 *)tmpbuf;
				gint16 *d = (gint16 *)audiobuf;
				for (int i=0; i < APLAY_FRAME_SIZE; i+=4) {
						*d++ = *s;
						*d++ = *s++; // Copy twice
				}
			}
	} else if (data->format.channels == 2) {
		if (fread(audiobuf, APLAY_FRAME_SIZE, 1, data->wav_fd) != 1) {
				pthread_mutex_unlock(&obj->object_mutex);
				return 0;
		}
	} else {
			printf("Huh? More than 2 channels?\n");
			exit(3);
	}
	if (obj->write_buffer) {
			obj->write_buffer->WriteSamples(audiobuf, APLAY_FRAME_SIZE >> 2);
	}
	pthread_mutex_unlock(&obj->object_mutex);
	return 1;
}


static int aplay_frame_seek(input_object *obj, int frame)
{
	int result = 0;

	if (!obj)
			return result;
	pthread_mutex_lock(&obj->object_mutex);
	struct aplay_local_data *data = (struct aplay_local_data *)obj->local_data;
	if (!data) {
			pthread_mutex_unlock(&obj->object_mutex);
			return result;
	}		
	long position = (frame * (APLAY_FRAME_SIZE >> (2 - data->format.channels)));
	position += data->header_size;
	result = !fseek(data->wav_fd, position, SEEK_SET);
	pthread_mutex_unlock(&obj->object_mutex);
	return result;
}


static int aplay_frame_size(input_object *obj)
{
	return APLAY_FRAME_SIZE;
}


static int aplay_nr_frames(input_object *obj)
{
	int result = 0;	
	if (!obj)
			return 0;
	pthread_mutex_lock(&obj->object_mutex);
	struct aplay_local_data *data = (struct aplay_local_data *)obj->local_data;
	if (!data) {
			pthread_mutex_unlock(&obj->object_mutex);
			return result;
	}		
	if (data->count)
			 result = (data->count / (APLAY_FRAME_SIZE >> ((2 - data->format.channels) + (data->format.bits == 8 ? 1 : 0))));
	pthread_mutex_unlock(&obj->object_mutex);
	return result;
}
		

static int aplay_sample_rate(input_object *obj)
{
	int result = 0;	
	if (!obj)
			return result;
	pthread_mutex_lock(&obj->object_mutex);
	struct aplay_local_data *data = (struct aplay_local_data *)obj->local_data;
	if (!data) {
			pthread_mutex_unlock(&obj->object_mutex);
			return result;
	}		
	result = data->format.rate;
	pthread_mutex_unlock(&obj->object_mutex);
	return result;
}


static int aplay_channels(input_object *obj)
{
	return 2; // Yes, always stereo, we take care of mono -> stereo converion
}


static unsigned long aplay_frame_to_sec(input_object *obj, int frame)
{
	unsigned long result = 0;
	if (!obj)
			return result;
	pthread_mutex_lock(&obj->object_mutex);
	struct aplay_local_data *data = (struct aplay_local_data *)obj->local_data;	
	if (!data) {
			pthread_mutex_unlock(&obj->object_mutex);
			return result;
	}		
	unsigned long pos = frame * (APLAY_FRAME_SIZE >> (2 - data->format.channels));
	result = (pos / (data->format.rate / 100) / data->format.channels / 2);
	pthread_mutex_unlock(&obj->object_mutex);
	return result;
}



static float aplay_test_support(char *name)
{
	WaveHeader wp;
	FILE *fd;

	if ((fd = fopen(name, "r")) == NULL) {
		return 0.0;
	}	
	if (fread(&wp, sizeof(WaveHeader), 1, fd) != 1) {
			fclose(fd);
			return 0.0;
	}
	fclose(fd);
    if (wp.main_chunk == WAV_RIFF && wp.chunk_type == WAV_WAVE &&
        wp.sub_chunk == WAV_FMT && wp.data_chunk == WAV_DATA) {
		if (wp.format != WAV_PCM_CODE) {
			fprintf(stderr, "APLAY: can't play not PCM-coded WAVE-files\n");
			return 0.0;
		}
	}
	return 1.0;
}

static int aplay_stream_info(input_object *obj, struct stream_info *info)
{
	if (!obj || !info)
			return 0;
	pthread_mutex_lock(&obj->object_mutex);
	struct aplay_local_data *data = (struct aplay_local_data *)obj->local_data;
	if (!data) {
			pthread_mutex_unlock(&obj->object_mutex);
			return 0;
	}		
	sprintf(info->stream_type, "%d-bit %dKhz %s WAV", 16, 
		data->format.rate / 1000, data->format.channels == 2 ? "stereo" : "mono");
	info->author[0] = 0;
	strcpy(info->title, data->wav_name);	
	pthread_mutex_unlock(&obj->object_mutex);
	
	return 1;
}

static int aplay_init()
{
	return 1;
}


input_plugin aplay_plugin = {
	{ "aplay WAV player  v1.0" },			 // Plugin name
	{ "Jaroslav Kysela and Andy Lo A Foe" }, // Author name
	aplay_init,
	aplay_test_support,
	aplay_open,
	aplay_close,
	aplay_play_frame,
	aplay_frame_seek,
	aplay_frame_size,
	aplay_nr_frames,
	aplay_frame_to_sec,
	aplay_sample_rate,
	aplay_channels,
	aplay_stream_info,
	NULL,
	NULL
};
